#include <ftw.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/xattr.h>
#include <sys/stat.h>
#include <errno.h>
#include <libgen.h>

#define MAX_PATH 4096
#define MAX_XATTR_SIZE 65536

void log_info(const char* message) {
	printf("INFO: %s\n", message);
}

void die(const char* message) {
	perror(message);
	exit(EXIT_FAILURE);
}

static int remove_callback(const char *fpath, const struct stat *sb, int typeflag, struct FTW *ftwbuf) {
	int ret = remove(fpath);
	if (ret) perror(fpath);
	return ret;
}

int cleanup_dir(const char* path) {
	// Use nftw for recursive deletion
	int ret = nftw(path, remove_callback, 64, FTW_DEPTH | FTW_PHYS);
	if (ret) perror("nftw cleanup failed");
	return ret;
}

// reference_path is path to which to create relative to
// absolute path is the full path of the file
// Inspired from https://stackoverflow.com/questions/36173695/how-to-retrieve-filepath-relatively-to-a-given-directory-in-c
char* relative_path(const char* reference_path, const char* absolute_path) {
	static char relative_path[MAX_PATH];
	// make copies to prevent altering original strings
	char* path_a = strdup(absolute_path);
	char* path_r = strdup(reference_path);
	int inc;
	int size_a = strlen(path_a)+1;
	int size_r = strlen(path_r)+1;

	for(inc = 0; inc < size_a && inc < size_r; inc += strlen(path_a+inc)+1) {
		char* token_a = strchr(path_a+inc, '/');
		char* token_r = strchr(path_r+inc, '/');
		if(token_a) token_a[0] = '\0';
		if(token_r) token_r[0] = '\0';
		if(strcmp(path_a+inc, path_r+inc) != 0) break;
	}

	for(int inc_r = inc; inc_r < size_r; inc_r += strlen(path_r+inc_r)+1) {
		if( !strchr(reference_path+inc_r, '/') ) break;
		strcat(relative_path, "..");
		strcat(relative_path, "/");
	}

	if(inc < size_a) strcat(relative_path, absolute_path+inc);

	return relative_path;
}

void get_xattrs(const char* path, const char* attrs_file) {
	ssize_t list_len = listxattr(path, NULL, 0);
	if (list_len < 0) die("listxattr failed");

	char* list = malloc(list_len);
	if (!list) die("malloc failed");

	if (listxattr(path, list, list_len) < 0) {
		free(list);
		die("listxattr failed");
	}

	FILE* fp = fopen(attrs_file, "w");
	if (!fp) die("fopen failed");

	char* attr = list;
	while (attr < list + list_len) {
		ssize_t value_len = getxattr(path, attr, NULL, 0);
		if (value_len > 0) {
			char* value = malloc(value_len);
			if (value && getxattr(path, attr, value, value_len) > 0) {
				fprintf(fp, "%s=%s\n", attr, value);
			}
			free(value);
		}
		attr += strlen(attr) + 1;
	}

	fclose(fp);
	free(list);
}

void set_xattrs(const char* path, const char* attrs_file) {
	FILE* fp = fopen(attrs_file, "r");
	if (!fp) die("fopen failed");

	char line[MAX_XATTR_SIZE];
	while (fgets(line, sizeof(line), fp)) {
		char* nl = strchr(line, '\n');
		if (nl) *nl = '\0';
		char* eq = strchr(line, '=');
		if (eq) {
			*eq = '\0';
			if (setxattr(path, line, eq + 1, strlen(eq + 1), 0) < 0) {
				char failed_attr[MAX_XATTR_SIZE];
				snprintf(failed_attr, sizeof(failed_attr),
					 "setxattr failed: %s %s", line, eq + 1);
				perror(failed_attr);
			}
		}
	}

	fclose(fp);
}

void replace_directory_with_symlink(const char* src, const char* dst) {
	char srcdir[MAX_PATH];
	char tmpdir[MAX_PATH];
	char attrs_path[MAX_PATH];
	char target[MAX_PATH];
	char* src_copy = strdup(src);

	if (!src_copy) die("strdup failed");

	// Create temporary directory
	strncpy(tmpdir, src, sizeof(src));
	strncat(tmpdir, ".XXXXXX", sizeof(tmpdir)-sizeof(src));
	if (!mkdtemp(tmpdir)) die("mkdtemp failed");

	// Get source directory and target path
	snprintf(srcdir, sizeof(srcdir), "%s", dirname(src_copy));
	snprintf(target, sizeof(target), "%s", relative_path(srcdir, dst));
	snprintf(attrs_path, sizeof(attrs_path), "%s/attrs", tmpdir);

	log_info("Replacing directory with symlink");

	// Save extended attributes
	get_xattrs(src, attrs_path);

	// Try to move directory
	char tmpdir_path[MAX_PATH];
	snprintf(tmpdir_path, sizeof(tmpdir_path), "%s/dir", tmpdir);
	if (rename(src, tmpdir_path) < 0) {
		if (errno == EXDEV) {
			// Handle cross-device link case
			if (cleanup_dir(src))
				die("Failed to remove directory");
		} else {
			die("Rename failed");
		}
	}

	// Create symbolic link
	if (symlink(target, src) < 0)
		die("Symlink creation failed");

	// Restore extended attributes
	set_xattrs(src, attrs_path);

	// Cleanup
	if (cleanup_dir(tmpdir))
		die("Failed to remove temporary directory");

	free(src_copy);
}

int main(int argc, char** argv) {
	if (argc != 3) {
		fprintf(stderr, "Usage: %s <src> <dst>\n", argv[0]);
		return EXIT_FAILURE;
	}

	replace_directory_with_symlink(argv[1], argv[2]);
	return EXIT_SUCCESS;
}
