#!/bin/sh
# shellcheck disable=SC3043
# SC3043: using 'local'

set -eu

# Keep chroots used for testing, do not remove them after tests are done
KEEP_CHROOTS="${KEEP_CHROOTS:-0}"
# If unset, a new root will be created. This could cause issues if multiple
# tests (re)use the same root!
CHROOT_DIR="${CHROOT_DIR:-}"

if ! command -v jchroot >/dev/null; then
	echo "Missing required app: jchroot"
	exit 1
fi

# $1: chroot path
# $2...: packages
apk_del() {
	local root="$1"
	shift
	jchroot -U "$root" -- apk del -vv "$@" >/dev/null || ( echo "failed to del packages, check log!" && return 1 )
}

# $1: chroot path
# $2...: packages
apk_add() {
	local root="$1"
	shift
	jchroot -U "$root" -- apk add --allow-untrusted -vv "$@" >/dev/null || ( echo "failed to add packages, check log!" && return 1 )
}

# $1: root path
# $2...: repos
add_repos() {
	local root="$1"
	shift
	mkdir -p "$root/etc/apk"
	for repo in "$@"; do
		echo "$repo" >> "$root/etc/apk/repositories"
	done
}

# $1: root path
create_root() {
	local root="$1"
	add_repos "$root" http://dl-cdn.alpinelinux.org/alpine/edge/main \
		  http://dl-cdn.alpinelinux.org/alpine/edge/community \
		  http://dl-cdn.alpinelinux.org/alpine/edge/testing

	# FIXME: downloading apkindex + packages for every root created is wasteful,
	# should cache/reuse stuff more

	# For test data/artifacts
	mkdir -p "$root/data"

	# Set up apk3 logging
	mkdir -p "$root/var/log"
	touch "$root/var/log/apk.log"

	mkdir -p "$root/etc/apk/keys"
	cp /etc/resolv.conf "$root/etc/resolv.conf"

	cp /etc/apk/keys/* "$root/etc/apk/keys/"

	apk --root "$root" --verbose add --no-cache --initdb --usermode \
		apk-tools busybox alpine-baselayout >/dev/null
}

print_header() {
	cat <<- EOF
		#####################################
		## $1
		#####################################
	EOF
}

print_step() {
	echo "#### $1"
}

# $1: root path
merge_usr() {
	local root="$1"
	cp "$(dirname "$0")"/../merge-usr "$root/usr/bin/"
	cp "$(dirname "$0")"/../merge_usr_replace_dir "$root/usr/bin/"
	jchroot -U "$root" -- merge-usr --verbose
}

# $1: rootfs path
# $*: Either a single apk world file OR a space-separated list of packages
pre_merge_add() {
	local root="$1"
	shift
	local packages
	if [ -f "$(dirname "$0")/$1" ]; then
		packages="$(tr '\n' ' ' < "$(dirname "$0")/$1")"
	else
		packages="$*"
	fi

	# FIXME: if `alpine-base` is included in this list, then apk-tools v2 is pulled
	# in and this breaks building the chroot. Maybe scrub that pkg from the list
	# before installing?
	# shellcheck disable=SC2086
	apk_add "$root" $packages || return 1
}

run_test() {
	local root run_found
	local test="$1"

	[ ! -f "$test" ] && echo "ERROR: test not found: $test" && return 1

	print_header "Running test: $test"

	if [ -n "$CHROOT_DIR" ]; then
		root="$CHROOT_DIR"
	else
		root="$(mktemp -d -p /tmp merge-usr."$(basename "$test")".XXXXXXXX)"
		[ -d "$root" ] || return 1
		print_step " ⛰️ Creating test root"
		if ! create_root "$root"; then
			echo "FAIL: unable to create testing rootfs"
			rm -rf "$root"
			return 1
		fi
	fi

	exec 4> /dev/null
	while IFS="" read -r ln; do
		case "$ln" in
		"@PRE_MERGE_PKGS_ADD "*)
			# Installed before test
			print_step " 📦 Installing packages pre-merge"
			# FIXME: this doesn't actually exit/bail...
			# shellcheck disable=SC2086
			pre_merge_add "$root" ${ln#* } || { echo "❌ Unable to run test!" && exit 1; }
			;;
		"@PRE_MERGE_REPOS_ADD "*)
			# Installed before test
			print_step " 📦 Adding repositories pre-merge"

			# shellcheck disable=SC2086
			add_repos "$root" ${ln#* } || { echo "❌ Unable to run test!" && exit 1; }
			;;
		"@PRE_CREATE_FILES "*)
			echo "${ln#* }" >> "$root/data/create_files"
			;;
		"@POST_MERGE_PKGS_ADD "*)
			echo "${ln#* }" >> "$root/data/add_pkgs"
			;;
		"@POST_MERGE_PKGS_DEL "*)
			echo "${ln#* }" >> "$root/data/del_pkgs"
			;;
		"@EXPECT")
			exec 4> "$root/data/expected"
			run_found=yes
			;;
		"@"*)
			echo "$test: invalid spec: $ln"
			run_found=""
			break
			;;
		*)
			echo "$ln" >&4
			;;
		esac
	done < "$test"
	exec 4> /dev/null

	# Create any files necessary for the test
	if [ -f "$root/data/create_files" ]; then
		print_step " 📁 Creating pre-merge files in root"
		while IFS="" read -r ln; do
			ln="${ln#* }"
			[ -n "$ln" ] || continue
			echo "Creating $ln"
			local path
			path="$root/$ln"
			if ! ( mkdir -p "$(dirname "$path")" && touch "$path" ); then
			 	echo "❌ Unable to run test!"
			 	return 1
			fi
		done < "$root/data/create_files"
	fi

	print_step " 💃 Doing the usr merge"
	if ! merge_usr "$root"; then
		echo "FAIL: unable to do the usr merge"
		[ "$KEEP_CHROOTS" -ne 1 ] && rm -rf "$root"
		return 1
	fi

	retcode=1
	if [ "$run_found" = "yes" ]; then
		# If removing packages is part of the test...
		if [ -f "$root/data/del_pkgs" ]; then
			print_step " 💥 Removing packages post-merge"
			while IFS="" read -r ln; do
				# FIXME: this is a test case, should mention in output + signal failure if appropriate
				# shellcheck disable=SC2086
				apk_del "$root" ${ln#* }
			done < "$root/data/del_pkgs"
		fi

		# If installing packages is part of the test...
		if [ -f "$root/data/add_pkgs" ]; then
			print_step " 📦 Installing packages post-merge"
			while IFS="" read -r ln; do
				# FIXME: this is a test case, should mention in output + signal failure if appropriate
				# shellcheck disable=SC2086
				apk_add "$root" ${ln#* }
			done < "$root/data/add_pkgs"
		fi

		# Check symlink expectations
		print_step "🔗 Checking files and symlinks"
		while IFS="" read -r ln; do
			local src target
			src="$(echo "$ln" | cut -d':' -f1)"
			target="$(echo "$ln" | cut -d':' -f2)"
			{ [ -z "$src" ] || [ -z "$target" ]; } && continue

			local result=""
			if [ "$src" = "$target" ]; then
				# No target separate target given, so check that source is a valid file
				printf "Check file: %-53s" "$src"
				if ! jchroot -U "$root" -- sh -c "[ -f $src ] || exit 1"; then
					result="File does not exist or is invalid: $src"
				fi
			else
				printf "Check symlink: %-50s" "$src --> $target"
				local realpath
				realpath="$(jchroot -U "$root" -- realpath "$src" 2>/dev/null)"
				if [ "$realpath" != "$target" ]; then
					result="Realpath of $src: $realpath"
				fi
			fi

			if [ -n "$result" ]; then
				echo  "❌"
				echo "$result"
				fail=$((fail+1))
				echo "FAIL: $test"
			else
				echo  "✅"
				retcode=0
			fi

		done < "$root/data/expected"
	fi

	du -hs "$root"
	mkdir -p "$(dirname "$0")"/logs
	cp "$root/var/log/apk.log" "$(dirname "$0")/logs/$(basename "$test")"

	[ "$KEEP_CHROOTS" -ne 1 ] && rm -rf "$root"
	return $retcode
}

TEST_TO_RUN="$*"

fail=0
pass=0
for test in ${TEST_TO_RUN:-"$(dirname "$0")"/*.test}; do
	if run_test "$test"; then
		pass=$((pass+1))
	else
		fail=$((fail+1))
	fi
done

if [ -z "$TEST_TO_RUN" ]; then
	total=$((fail+pass))
	if [ "$fail" != "0" ]; then
		echo "FAIL: $fail of $total test cases failed"
	else
		echo "OK: all $total test cases passed"
	fi
fi

[ "$fail" -eq 0 ] || exit 1
exit 0
